@@ -44,87 +44,93 @@ defmodule ExUnit.Diff do
 
   # Main entry point for recursive diff
 
-  defp diff(left, right, %{context: :match} = env), do: diff_quoted(left, right, env)
+  defp diff(left, right, %{context: :match} = env) do
+    case left do
+      {_, [original: original] ++ _, _} ->
+        diff_quoted(original, right, left, env)
+
+      _ ->
+        diff_quoted(left, right, nil, env)
+    end
+  end
+
   defp diff(left, right, env), do: diff_value(left, right, env)
 
   # diff quoted
 
-  defp diff_quoted({:_, _, context} = left, right, env) when is_atom(context) do
+  defp diff_quoted({:_, _, context} = left, right, _expanded, env) when is_atom(context) do
     diff_right = escape(right)
     diff = %__MODULE__{equivalent?: true, left: left, right: diff_right}
     {diff, env}
   end
 
-  defp diff_quoted({:^, _, [{name, _, context}]} = left, right, env)
+  defp diff_quoted({:^, _, [{name, _, context}]} = left, right, _expanded, env)
        when is_atom(name) and is_atom(context) do
     diff_pin(left, right, env)
   end
 
-  defp diff_quoted({name, _, context} = left, right, env)
+  defp diff_quoted({name, _, context} = left, right, _expanded, env)
        when is_atom(name) and is_atom(context) and
               name not in [:__MODULE__, :__DIR__, :__STACKTRACE__, :__ENV__, :__CALLER__] do
     diff_var(left, right, env)
   end
 
-  defp diff_quoted({:-, _, [number]}, right, env) when is_number(number) do
-    diff_quoted(-number, right, env)
+  defp diff_quoted({:-, _, [number]}, right, _expanded, env) when is_number(number) do
+    diff_quoted(-number, right, nil, env)
   end
 
-  defp diff_quoted({:+, _, [number]}, right, env) when is_number(number) do
-    diff_quoted(number, right, env)
+  defp diff_quoted({:+, _, [number]}, right, _expanded, env) when is_number(number) do
+    diff_quoted(number, right, nil, env)
   end
 
-  defp diff_quoted({:++, meta, [prefix, suffix]}, right, env) when is_list(right) do
-    case prefix do
-      {_, [expanded: expanded] ++ _, _} ->
-        diff_maybe_improper_list({:++, meta, [expanded, suffix]}, right, env)
-
-      _ ->
-        diff_maybe_improper_list({:++, meta, [prefix, suffix]}, right, env)
-    end
+  defp diff_quoted({:++, meta, [prefix, suffix]}, right, _expanded, env) when is_list(right) do
+    diff_maybe_improper_list({:++, meta, [prefix, suffix]}, right, env)
   end
 
-  defp diff_quoted({:{}, _, left}, right, env) when is_tuple(right) do
+  defp diff_quoted({:{}, _, left}, right, _expanded, env) when is_tuple(right) do
     diff_tuple(left, Tuple.to_list(right), env)
   end
 
-  defp diff_quoted({_, _} = left, right, env) when is_tuple(right) do
+  defp diff_quoted({_, _} = left, right, _expanded, env) when is_tuple(right) do
     diff_tuple(Tuple.to_list(left), Tuple.to_list(right), env)
   end
 
-  defp diff_quoted({:%, _, [struct, {:%{}, _, kw}]}, %{} = right, env) when is_list(kw) do
+  defp diff_quoted({:%, _, [struct, {:%{}, _, kw}]}, %{} = right, _expanded, env)
+       when is_list(kw) do
     diff_quoted_struct([__struct__: struct] ++ kw, right, env)
   end
 
-  defp diff_quoted({:%{}, _, kw}, %{} = right, env) when is_list(kw) do
+  defp diff_quoted({:%{}, _, kw}, %{} = right, _expanded, env) when is_list(kw) do
     diff_quoted_struct(kw, right, env)
   end
 
-  defp diff_quoted({:<>, _, [literal, _]} = left, right, env)
+  defp diff_quoted({:<>, _, [literal, _]} = left, right, _expanded, env)
        when is_binary(literal) and is_binary(right) do
     diff_string_concat(left, right, env)
   end
 
-  defp diff_quoted({:when, _, [_, _]} = left, right, env) do
+  defp diff_quoted({:when, _, [_, _]} = left, right, _expanded, env) do
     diff_guard(left, right, env)
   end
 
-  defp diff_quoted({_, [{:expanded, expanded} | _], _} = left, right, env) do
-    macro = Macro.update_meta(left, &Keyword.delete(&1, :expanded))
-    diff_macro(macro, expanded, right, env)
+  defp diff_quoted({_, _, _} = left, right, expanded, env) when expanded != nil do
+    expanded = Macro.update_meta(expanded, &Keyword.delete(&1, :original))
+    {diff, post_env} = diff(expanded, right, env)
+    diff_left = update_diff_meta(left, !diff.equivalent?)
+    {%{diff | left: diff_left}, post_env}
   end
 
-  defp diff_quoted(left, right, env) when is_list(left) and is_list(right) do
+  defp diff_quoted(left, right, _expanded, env) when is_list(left) and is_list(right) do
     diff_maybe_list(left, right, env)
   end
 
-  defp diff_quoted(left, right, env)
+  defp diff_quoted(left, right, _expanded, env)
        when is_atom(left) or is_number(left) or is_reference(left) or
               is_pid(left) or is_function(left) or is_binary(left) do
     diff_value(left, right, env)
   end
 
-  defp diff_quoted(left, right, %{context: :match} = env) do
+  defp diff_quoted(left, right, _expanded, %{context: :match} = env) do
     diff_left = update_diff_meta(left, true)
     diff_right = escape(right) |> update_diff_meta(true)
     diff = %__MODULE__{equivalent?: false, left: diff_left, right: diff_right}
@@ -174,25 +180,45 @@ defmodule ExUnit.Diff do
     diff_string(left, right, ?", env)
   end
 
+  defp diff_value(left, right, env) when is_function(left) and is_function(right) do
+    # If they are local functions in the same module and same name,
+    # then they have different environment, which we compare.
+    if Function.info(left, :type) == {:type, :local} and
+         Function.info(right, :type) == {:type, :local} and
+         Function.info(left, :module) == Function.info(right, :module) and
+         Function.info(left, :new_uniq) == Function.info(right, :new_uniq) and
+         Function.info(left, :new_index) == Function.info(right, :new_index) do
+      {:env, left_env} = Function.info(left, :env)
+      {:env, right_env} = Function.info(right, :env)
+      {diff, env} = diff_maybe_improper_list(left_env, right_env, env)
+
+      diff = %{
+        diff
+        | left: %{custom: &function_env_to_algebra(left, diff.left, &1)},
+          right: %{custom: &function_env_to_algebra(right, diff.right, &1)}
+      }
+
+      {diff, env}
+    else
+      non_recursive_diff_value(left, right, env)
+    end
+  end
+
   defp diff_value(left, right, env) do
+    non_recursive_diff_value(left, right, env)
+  end
+
+  defp non_recursive_diff_value(left, right, env) do
     diff_left = escape(left) |> update_diff_meta(true)
     diff_right = escape(right) |> update_diff_meta(true)
     diff = %__MODULE__{equivalent?: false, left: diff_left, right: diff_right}
     {diff, env}
   end
 
-  # Macros
-
-  defp diff_macro(macro, expanded, right, env) do
-    {diff, post_env} = diff(expanded, right, env)
-    diff_left = update_diff_meta(macro, !diff.equivalent?)
-    {%{diff | left: diff_left}, post_env}
-  end
-
   # Guards
 
   defp diff_guard({:when, _, [expression, clause]}, right, env) do
-    {diff_expression, post_env} = diff_quoted(expression, right, env)
+    {diff_expression, post_env} = diff_quoted(expression, right, nil, env)
 
     {guard_clause, guard_equivalent?} =
       if diff_expression.equivalent? do
@@ -211,52 +237,59 @@ defmodule ExUnit.Diff do
     {diff, post_env}
   end
 
-  defp diff_guard_clause({op, _, [clause1, clause2]}, bindings) when op in [:when, :or, :and] do
-    {diff_clause1, clause1_equivalent?} = diff_guard_clause(clause1, bindings)
-    {diff_clause2, clause2_equivalent?} = diff_guard_clause(clause2, bindings)
-
-    equivalent? =
-      case op do
-        :and -> clause1_equivalent? and clause2_equivalent?
-        _other -> clause1_equivalent? or clause2_equivalent?
-      end
-
-    diff = {op, [], [diff_clause1, diff_clause2]}
-    {diff, equivalent?}
-  end
-
   defp diff_guard_clause(quoted, bindings) do
-    expanded =
-      Macro.prewalk(quoted, fn
-        {_, [{:expanded, expanded} | _], _} ->
-          expanded
+    case original_or_current(quoted) do
+      {op, _, [clause1, clause2]} when op in [:and, :or, :when] ->
+        {diff_clause1, clause1_equivalent?} = diff_guard_clause(clause1, bindings)
+        {diff_clause2, clause2_equivalent?} = diff_guard_clause(clause2, bindings)
 
-        {var, _, context} = expr when is_atom(var) and is_atom(context) ->
-          if Map.has_key?(bindings, var_context(expr)) do
-            expr
-          else
-            throw(:abort)
+        equivalent? =
+          case op do
+            :and -> clause1_equivalent? and clause2_equivalent?
+            _other -> clause1_equivalent? or clause2_equivalent?
           end
 
-        other ->
-          other
-      end)
+        diff = {op, [], [diff_clause1, diff_clause2]}
+        {diff, equivalent?}
 
-    {equivalent?, _bindings} = Code.eval_quoted(expanded, Map.to_list(bindings))
-    {update_diff_meta(quoted, !equivalent?), equivalent?}
-  catch
-    :abort -> {quoted, false}
+      _ ->
+        {original, valid?} =
+          Macro.postwalk(quoted, true, fn
+            {_, [original: original] ++ _, _}, valid? ->
+              {original, valid?}
+
+            {var, _, context} = expr, valid? when is_atom(var) and is_atom(context) ->
+              if Map.has_key?(bindings, var_context(expr)) do
+                {expr, valid?}
+              else
+                {expr, false}
+              end
+
+            other, valid? ->
+              {other, valid?}
+          end)
+
+        if valid? do
+          {equivalent?, _bindings} = Code.eval_quoted(quoted, Map.to_list(bindings))
+          {update_diff_meta(original, equivalent? != true), equivalent? == true}
+        else
+          {original, false}
+        end
+    end
   end
 
+  defp original_or_current({_, [original: original] ++ _, _}), do: original
+  defp original_or_current(current), do: current
+
   # Pins
 
   defp diff_pin({:^, _, [var]} = pin, right, %{pins: pins} = env) do
     identifier = var_context(var)
     %{^identifier => pin_value} = pins
-    {diff, post_env} = diff_value(pin_value, right, env)
+    {diff, post_env} = diff_value(pin_value, right, %{env | context: :===})
 
     diff_left = update_diff_meta(pin, not diff.equivalent?)
-    {%{diff | left: diff_left}, post_env}
+    {%{diff | left: diff_left}, %{post_env | context: :match}}
   end
 
   # Vars
@@ -318,7 +351,15 @@ defmodule ExUnit.Diff do
 
   defp diff_maybe_list(left, right, env) do
     if List.ascii_printable?(left) and List.ascii_printable?(right) do
-      diff_string(List.to_string(left), List.to_string(right), ?', env)
+      {diff, env} = diff_string(List.to_string(left), List.to_string(right), ?", env)
+
+      diff = %{
+        diff
+        | left: %{custom: &charlist_to_algebra(diff.left, &1)},
+          right: %{custom: &charlist_to_algebra(diff.right, &1)}
+      }
+
+      {diff, env}
     else
       diff_maybe_improper_list(left, right, env)
     end
@@ -708,7 +749,7 @@ defmodule ExUnit.Diff do
   defp load_struct(struct) do
     if is_atom(struct) and struct != nil and
          Code.ensure_loaded?(struct) and function_exported?(struct, :__struct__, 0) do
-      struct.__struct__
+      struct.__struct__()
     end
   end
 
@@ -790,7 +831,6 @@ defmodule ExUnit.Diff do
       merge_diff(parsed_diff, quoted_diff, fn left1, left2, right1, right2 ->
         new_left = rebuild_concat_string(left1, left2, indexes)
         new_right = rebuild_split_strings(right1, right2)
-
         {new_left, new_right}
       end)
 
@@ -820,12 +860,15 @@ defmodule ExUnit.Diff do
     left
   end
 
-  defp rebuild_split_strings({:__block__, meta, left_list}, {:__block__, _, right_list}) do
-    {:__block__, meta, left_list ++ right_list}
+  defp rebuild_split_strings(
+         %{contents: left, delimiter: delimiter},
+         %{contents: right, delimiter: delimiter}
+       ) do
+    %{contents: left ++ right, delimiter: delimiter}
   end
 
-  defp rebuild_split_strings({:__block__, meta, left_list}, right) do
-    {:__block__, meta, left_list ++ [right]}
+  defp rebuild_split_strings(%{contents: contents, delimiter: delimiter}, right) do
+    %{contents: contents ++ [{false, right}], delimiter: delimiter}
   end
 
   defp rebuild_concat_string(literal, nil, []) do
@@ -839,58 +882,53 @@ defmodule ExUnit.Diff do
   defp rebuild_concat_string(literal, quoted, [index | rest]) do
     {next, continue} = next_concat_result(literal, index)
     rebuilt_right = rebuild_concat_string(continue, quoted, rest)
-
     {:<>, [], [next, rebuilt_right]}
   end
 
-  defp next_concat_result({:__block__, [{:delimiter, _} | _] = meta, list}, index) do
-    {next, continue} = next_concat_result(list, index)
-    {{:__block__, meta, next}, {:__block__, meta, continue}}
+  defp next_concat_result(%{contents: contents, delimiter: delimiter}, index) do
+    {next, continue} = next_concat_result(contents, index)
+    {%{contents: next, delimiter: delimiter}, %{contents: continue, delimiter: delimiter}}
   end
 
-  defp next_concat_result([head | tail], index) do
-    {string, diff_meta?} = extract_diff_meta(head)
-    length = String.length(string)
+  defp next_concat_result([{diff?, head} | tail], index) do
+    length = String.length(head)
 
     cond do
       length > index ->
-        {next, continue} = String.split_at(string, index)
-        next = [update_diff_meta(next, diff_meta?)]
-        continue = [update_diff_meta(continue, diff_meta?) | tail]
-
-        {next, continue}
+        {next, continue} = String.split_at(head, index)
+        {[{diff?, next}], [{diff?, continue} | tail]}
 
       length < index ->
         {next, continue} = next_concat_result(tail, index - length)
-        {[head | next], continue}
+        {[{diff?, head} | next], continue}
 
       true ->
-        {[head], tail}
+        {[{diff?, head}], tail}
     end
   end
 
-  defp block_delimiter(contents, nil),
-    do: {:__block__, [], contents}
-
-  defp block_delimiter(contents, container),
-    do: {:__block__, [delimiter: container], contents}
-
   defp string_script_to_diff([], delimiter, equivalent?, left, right) do
-    left = block_delimiter(Enum.reverse(left), delimiter)
-    right = block_delimiter(Enum.reverse(right), delimiter)
+    left = %{delimiter: delimiter, contents: Enum.reverse(left)}
+    right = %{delimiter: delimiter, contents: Enum.reverse(right)}
     %__MODULE__{equivalent?: equivalent?, left: left, right: right}
   end
 
   defp string_script_to_diff([{:eq, string} | tail], delimiter, equivalent?, left, right) do
-    string_script_to_diff(tail, delimiter, equivalent?, [string | left], [string | right])
+    string_script_to_diff(
+      tail,
+      delimiter,
+      equivalent?,
+      [{false, string} | left],
+      [{false, string} | right]
+    )
   end
 
   defp string_script_to_diff([{:del, string} | tail], delimiter, _equivalent?, left, right) do
-    string_script_to_diff(tail, delimiter, false, [update_diff_meta(string, true) | left], right)
+    string_script_to_diff(tail, delimiter, false, [{true, string} | left], right)
   end
 
   defp string_script_to_diff([{:ins, string} | tail], delimiter, _equivalent?, left, right) do
-    string_script_to_diff(tail, delimiter, false, left, [update_diff_meta(string, true) | right])
+    string_script_to_diff(tail, delimiter, false, left, [{true, string} | right])
   end
 
   # Numbers
@@ -908,21 +946,6 @@ defmodule ExUnit.Diff do
     wrap_on_diff(quoted, &safe_to_algebra/2, diff_wrapper)
   end
 
-  defp safe_to_algebra({:__block__, meta, list}, diff_wrapper) do
-    content_docs =
-      for item <- list do
-        # Each element of the list is either a literal or a wrapped literal in a diff
-        wrap_on_diff(item, fn literal, _ -> literal end, diff_wrapper)
-      end
-
-    if container = meta[:delimiter] do
-      delimiter = to_string([container])
-      Algebra.concat([delimiter] ++ content_docs ++ [delimiter])
-    else
-      Algebra.concat(content_docs)
-    end
-  end
-
   defp safe_to_algebra(list, diff_wrapper) when is_list(list) do
     container_to_algebra("[", list, "]", diff_wrapper, select_list_item_algebra(list))
   end
@@ -964,6 +987,26 @@ defmodule ExUnit.Diff do
     inspect(escaped)
   end
 
+  # Custom encoding for delimiters+contents
+  defp safe_to_algebra(%{delimiter: delimiter, contents: contents}, diff_wrapper) do
+    content_docs =
+      for {diff?, content} <- contents do
+        if diff?, do: diff_wrapper.(content), else: content
+      end
+
+    if delimiter do
+      delimiter = List.to_string([delimiter])
+      Algebra.concat([delimiter, Algebra.concat(content_docs), delimiter])
+    else
+      Algebra.concat(content_docs)
+    end
+  end
+
+  # Custom encoding for functions
+  defp safe_to_algebra(%{custom: literal}, diff_wrapper) do
+    literal.(diff_wrapper)
+  end
+
   defp safe_to_algebra(literal, _diff_wrapper) do
     inspect(literal)
   end
@@ -1052,17 +1095,46 @@ defmodule ExUnit.Diff do
     end
   end
 
+  defp charlist_to_algebra(%{contents: contents}, diff_wrapper) do
+    content_doc =
+      for {diff, content} <- contents do
+        if diff, do: diff_wrapper.(content), else: content
+      end
+
+    Algebra.concat(["~c\"", Algebra.concat(content_doc), "\""])
+  end
+
+  defp function_env_to_algebra(function, ast, diff_wrapper) do
+    "#Function<" <> contents = inspect(function)
+
+    Algebra.concat([
+      "#Function<",
+      Algebra.nest(
+        Algebra.concat([
+          Algebra.line(),
+          binary_slice(contents, 0..-2//1),
+          Algebra.line(),
+          to_algebra(ast, diff_wrapper)
+        ]),
+        2
+      ),
+      Algebra.line(),
+      ">"
+    ])
+  end
+
   # Diff helpers
 
   defp add_hint(%{hints: hints} = env, hint) do
     if hint in hints, do: env, else: %{env | hints: [hint | hints]}
   end
 
-  # The left side is only escaped if it is a value
   defp maybe_escape(other, %{context: :match}), do: other
   defp maybe_escape(other, _env), do: escape(other)
 
-  # We escape it by wrapping it in one element tuple which is not valid AST
+  # We escape container types to make a distinction between AST
+  # and values that should be inspected. All other values have no
+  # special AST representation, so we can keep them as is.
   defp escape(other) when is_list(other) or is_tuple(other), do: {other}
   defp escape(other), do: other
 
